// file: $isip/class/algo/Cepstrum/cep_03.cc
// version: $Id: cep_03.cc,v 1.11 2002/07/17 21:59:45 parihar Exp $
//

// isip include files
//
#include "Cepstrum.h"

//------------------------------------------------------------------
// these methods have to be in the same file so they can use the same
// static parser pointer
//------------------------------------------------------------------

// declare a static sof parser pointer
//
static SofParser* parser_d = (SofParser*)NULL;

// method: setParser
//
// arguments:
//  SofParser* parser: (input) sof file object
//
// return: a boolean value indicating status
//
// this method sets the parser from the parent object to be used in
// the next readData method.
//
boolean Cepstrum::setParser(SofParser* parser_a) {

  // set the parser
  //
  parser_d = parser_a;

  // exit gracefully
  //
  return true;
}

// method: read
//
// arguments:
//  Sof& sof: (input) sof file object
//  long tag: (input) sof object instance tag
//  const String& name: (input) sof object instance name
//
// return: a boolean value indicating status
//
// this method has the object, with specified name and tag, read
// itself from an Sof file
//
boolean Cepstrum::read(Sof& sof_a, long tag_a, const String& name_a) {

  // read the instance of the object from the Sof file
  //
  if (!sof_a.find(name_a, tag_a)) {
    return false;
  }
  
  // read the actual data from the sof file
  //
  if (!readData(sof_a)) {
    return false;
  }
  
  // exit gracefully
  //
  return true;  
}

// method: readData
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//  long size: (input) size in bytes of object (or full_size)
//  boolean param: (input) is the parameter name in the file?
//  boolean nested: (input) are we nested?
//
// return: a boolean indicating status
//
// this method has the object read itself from an Sof file. it assumes
// that the Sof file is already positioned correctly.
//
boolean Cepstrum::readData(Sof& sof_a, const String& pname_a,
			   long size_a, boolean param_a,
			   boolean nested_a) {

  // declare local variables
  //
  boolean status = false;
  
  // we need a parser
  // 
  if (parser_d == (SofParser*)NULL) {

    // initialize the parser
    //
    parser_d = new SofParser;

    // check if the object is nested
    //
    if (nested_a) {
      parser_d->setNest();
    }
    
    // load the parser
    //
    if (!parser_d->load(sof_a, size_a)) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::READ,
                           __FILE__, __LINE__, Error::WARNING);
    }           
  }

  // set the debug level of the parser
  //
  parser_d->setDebug(debug_level_d.getIndex());

  // get the algorithm
  //
  if (parser_d->isPresent(sof_a, PARAM_ALGORITHM)) {
    if (!ALGO_MAP.readElementData((long&)algorithm_d, sof_a, PARAM_ALGORITHM,
				  parser_d->getEntry(sof_a,
						     PARAM_ALGORITHM))) {
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::READ, __FILE__,
			   __LINE__, Error::WARNING);
    }
  }
  else {
    algorithm_d = DEF_ALGORITHM;
  }

  // read the debug level
  //
  if (parser_d->isPresent(sof_a, PARAM_DBGL)) {
    if (!debug_level_d.readData(sof_a, PARAM_DBGL,
				parser_d->getEntry(sof_a, PARAM_DBGL))) {
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    debug_level_d = Integral::DEF_DEBUG;
  }
  
  // check known algorithms: invoke the proper lower-level method
  //
  if ((algorithm_d == IDCT) || (algorithm_d == IDFT)) {
    status = readDataCommon(sof_a, pname_a, param_a, nested_a);
  }

  // check unsupported algorithms
  //
  else {
    return Error::handle(name(), L"readData",
                         ERR_UNKALG, __FILE__, __LINE__);
  }

  // check that all parameters are accounted for
  //
  if (!parser_d->checkParams(sof_a)) {
    
    // delete the parser
    //
    delete parser_d;
    parser_d = (SofParser*)NULL;

    // return a warning message
    //
    return Error::handle(name(), L"readData", Error::IO,
                         __FILE__, __LINE__, Error::WARNING);
  }
  
  // delete the parser
  //
  delete parser_d;
  parser_d = (SofParser*)NULL;
  
  // exit gracefully
  //
  return status;
}

// method: readDataCommon
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//  long size: (input) size in bytes of object (or full_size)
//  boolean param: (input) is the parameter name in the file?
//  boolean nested: (input) are we nested?
//
// return: a boolean value indicating status
//
// this method has the object read itself from an Sof file. it assumes
// that the Sof file is already positioned correctly. this method is
// called by the readData method. in algorithm classes, we have an
// enumeration which defines the algorithm. depending upon the
// requirements of those algorithms, we implement private
// algorithm-specific readData methods. the readData method reads the
// algorithm and calls those private readData methods depending upon
// the type of algorithm. in Cepstrum class, both the algorithms will
// read same objects and thus we have implemented readDataCommon
// method, which is a common read method for both algorithms.
//
boolean Cepstrum::readDataCommon(Sof& sof_a, const String& pname_a,
				 long size_a, boolean param_a,
				 boolean nested_a) {

  // get the implementation
  //
  if (parser_d->isPresent(sof_a, PARAM_IMPLEMENTATION)) {
    if (!IMPL_MAP.readElementData((long&)implementation_d, sof_a,
				  PARAM_IMPLEMENTATION,
				  parser_d->getEntry(sof_a,
						     PARAM_IMPLEMENTATION))) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    implementation_d = DEF_IMPLEMENTATION;
  }
  
  // read the order
  //
  if (parser_d->isPresent(sof_a, PARAM_ORDER)) {
    if (!order_d.readData(sof_a, PARAM_ORDER,
				     parser_d->getEntry(sof_a,
						     PARAM_ORDER),
				     false, false)) {
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;

      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    order_d.assign(DEF_ORDER);
  }

  // read the minimum amplitude
  //
  if (parser_d->isPresent(sof_a, PARAM_MIN_AMP)) {

    // set the flag if the minimum amplitude is defined
    //
    if (!min_amp_d.readData(sof_a, PARAM_MIN_AMP,
			    parser_d->getEntry(sof_a, PARAM_MIN_AMP),
			    false, false)) {
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    min_amp_d.assign(DEF_MIN_AMP);
  }
    
  // read the minimum amplitude flag
  //
  if (parser_d->isPresent(sof_a, PARAM_FLAG_MIN_AMP)) {

    // set the flag if the minimum amplitude is defined
    //
    if (!flag_min_amp_d.readData(sof_a, PARAM_FLAG_MIN_AMP,
			    parser_d->getEntry(sof_a, PARAM_FLAG_MIN_AMP),
				 false, false)) {
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    flag_min_amp_d.assign(DEF_FLAG_MIN_AMP);
  }
    
  // exit gracefully
  //
  return true;
}
