// file: $isip/class/math/scalar/MScalar/MScalar.h
// version: $Id: MScalar.h,v 1.86 2002/08/02 23:46:55 zheng Exp $
//

// make sure definitions are only made once
//
#ifndef ISIP_MSCALAR
#define ISIP_MSCALAR

// isip include files
//
#ifndef ISIP_MSCALAR_METHODS
#include "MScalarMethods.h"
#endif

#ifndef ISIP_STRING
#include <String.h>
#endif

#ifndef ISIP_RANDOM
#include <Random.h>
#endif

// MScalar: the scalar template class is inherited by other
// scalar classes. TIntegral is an Integral type, TSize is a size
// specific data type.  most of the methods in this class are inline
// wrappers to the MScalarMethods class functions. this is done to
// allow us to write one method per file, thus avoiding the compilation
// overhead inherent in compiling template classes with multiple
// explicit instantiations.
//
template<class TIntegral, class TSize>
class MScalar {

  //---------------------------------------------------------------------------
  //
  // public constants
  //
  //---------------------------------------------------------------------------
public:
  
  // define the class name
  //
  static const String CLASS_NAME;

  //----------------------------------------
  //
  // other important constants
  //
  //----------------------------------------
  
  // random number related constants
  //
  static const double RAND_BYTE_MAX = Integral::TWO_EXP8;
  static const double RAND_USHORT_MAX = Integral::TWO_EXP16;
  static const double RAND_ULONG_MAX = Integral::TWO_EXP32;
  static const double RAND_ULLONG_MAX = Integral::TWO_EXP64;
  static const double RAND_SHORT_MAX = Integral::TWO_EXP15;
  static const double RAND_LONG_MAX = Integral::TWO_EXP31;
  static const double RAND_LLONG_MAX = Integral::TWO_EXP63;

  // i/o related constants
  //
  static const String DEF_PARAM;
  
  //----------------------------------------
  //
  // default values and arguments
  //
  //----------------------------------------
  
  // define the default value(s) of the class data
  //
  static const TIntegral DEF_VALUE;

  //----------------------------------------
  //
  // error codes
  //
  //----------------------------------------  
  
  static const long ERR = 20000;

  //---------------------------------------------------------------------------
  //
  // protected data
  //
  //---------------------------------------------------------------------------
protected:

  // the container for the value of this item
  //
  TIntegral value_d;

  // declare a static debug level for all class instantiations:
  // we don't want every object to waste space on a debug level. let
  //  them share this across all classes.
  //
  static Integral::DEBUG debug_level_d;
  
  //---------------------------------------------------------------------------
  //
  // required public methods
  //
  //---------------------------------------------------------------------------
public:
  
  // method: name
  //
  static const String& name() {
    return CLASS_NAME;
  }
  
  // method: diagnose
  //
  static boolean diagnose(Integral::DEBUG debug_level) {
    return MScalarMethods::
      diagnose<MScalar<TIntegral, TSize>, TIntegral>(debug_level);
  }
 
  // method: setDebug
  //
  static boolean setDebug(Integral::DEBUG level) {
    debug_level_d = level;
    return true;
  }

  // other debug methods
  //
  boolean debug(const unichar* msg) const;
  
  // method: destructor
  //
  ~MScalar() {}
 
  // method: default constructor
  //
  MScalar() {}

  // method: copy constructor
  //
  MScalar(const MScalar& arg) {
    assign(arg);
  }

  // method: assign
  //
  boolean assign(const MScalar& arg) {
    value_d = (TIntegral)arg;
    return true;
  }

  // method operator=
  //  we need to have the overloaded operator '=' defined in the
  //  derived classes instead of the base template class. this is
  //  required since c++ predefines the operator '=' for any class by
  //  default and this hides any other definition of this operator in
  //  the base class.
  //

  // method: sofSize
  //
  long sofSize() const {
    return (long)sizeof(TSize);
  }
  
  // method: read
  //
  boolean read(Sof& sof, long tag, const String& name = CLASS_NAME) {
    return MScalarMethods::read(*this, sof, tag, name);
  }
  
  // method: write
  //
  boolean write(Sof& sof, long tag, const String& name = CLASS_NAME) const {
    return MScalarMethods::write(*this, sof, tag, name);
  }

  // method: readData
  //
  boolean readData(Sof& sof, const String& pname = DEF_PARAM,
			  long size = SofParser::FULL_OBJECT,
			  boolean param = true, boolean nested = false) {
    return MScalarMethods::readData(*this, sof, pname, size, param, nested);
  }

  // method: writeData
  //
  boolean writeData(Sof& sof, const String& pname = DEF_PARAM) const {
    return MScalarMethods::writeData(*this, sof, pname);
  }
  
  // method: eq
  //
  boolean eq(const MScalar& arg_a) const {
    return (value_d == arg_a.value_d);
  }

  // memory management methods:
  //  new and delete methods are omitted because they are defined in the
  //  classes that instantiate this template
  //

  // method: clear
  //
  boolean clear(Integral::CMODE ctype = Integral::DEF_CMODE) {
    value_d = (TIntegral)DEF_VALUE;
    return true;
  }
  
  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  extensions to required methods
  //
  //---------------------------------------------------------------------------

  // method: constructor
  //  note that this method is useful to have for the diagnose method.
  //  it is masked by higher-level classes that use this as a template
  //  and have their own version of this constructor.
  //
  MScalar(TIntegral arg) {
    assign(arg);
  }

  // method: assign
  //
  template<class TAIntegral>
  boolean assign(TAIntegral arg) {
    value_d = (TIntegral)arg;
    return true;
  }

  // method: assign
  //
  boolean assign(const String& arg) {
    arg.get(value_d);
    return true;
  }
  
  // method: get
  //
  boolean get(String& arg) const {
    return arg.assign(value_d);
  }

  // method: memSize
  //
  long memSize() const {
    return (long)sizeof(value_d);
  }

  // method: presentValue
  //  define functions which are used to wrap member functions that have
  //  different implementations for different data types, and return the
  //  correct value by using template specialization.
  //
  TIntegral presentValue(double arg) const {
    return (TIntegral)Integral::round(arg);
  }

  // method: presentValue
  //  the complex function overload is a dummy function. all of the
  //  instantiations are specialized.
  //
  TIntegral presentValue(complexdouble arg) const {
    return Error::handle(name(), L"presentValue", Error::TEMPLATE_TYPE,
			 __FILE__, __LINE__);
  }

  // method: almostEqual
  //
  boolean almostEqual(TIntegral arg) const {
    return Integral::almostEqual(value_d, arg);
  }

  // method: almostEqual
  //  
  boolean almostEqual(TIntegral arg, double percent,
		      double bound = Integral::DEF_BOUND) const {
    return Integral::almostEqual(value_d, arg, percent, bound);
  }

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  operator overloads
  //
  //---------------------------------------------------------------------------

  // method: operator TIntegral()
  //
  operator TIntegral() const {
    return value_d;
  }

  // method: operator +
  //
  TIntegral operator+ (TIntegral arg) const {
    return value_d + arg;
  }

  // method: operator -
  //
  TIntegral operator- (TIntegral arg) const {
    return value_d - arg;
  }

  // method: operator *
  //
  TIntegral operator* (TIntegral arg) const {
    return value_d * arg;
  }

  // method: operator /
  //
  TIntegral operator/ (TIntegral arg) const {
    return value_d / arg;
  }

  // method: operator %
  //
  TIntegral operator% (TIntegral arg) const {
    return value_d % arg;
  }

  // method: operator ++
  //
  TIntegral operator++ () {
    return ++value_d;
  }

  // method: operator --
  //
  TIntegral operator-- () {
    return --value_d;
  }
  
  // method: operator ++
  //
  TIntegral operator++ (int) {
    return value_d++;
  }
  
  // method: operator --
  //
  TIntegral operator-- (int) {
    return value_d--;
  }

  // method: operator +=
  //
  MScalar& operator+= (TIntegral arg) {
    value_d += arg;
    return *this;
  }

  // method: operator -=
  //
  MScalar& operator-= (TIntegral arg) {
    value_d -= arg;
    return *this;
  }

  // method: operator *=
  //
  MScalar& operator*= (TIntegral arg) {
    value_d *= arg;
    return *this;
  }

  // method: operator /=
  //
  MScalar& operator/= (TIntegral arg) {
    value_d /= arg;
    return *this;
  }

  // method: operator %=
  //
  MScalar& operator%= (TIntegral arg) {
    value_d %= arg;
    return *this;
  }
   
  // method: operator ==
  //
  boolean operator == (TIntegral arg) const {
    return eq(arg);
  }

  // method: operator !=
  //
  boolean operator != (TIntegral arg) const {
    return ne(arg);
  }

  // method: operator <
  //
  boolean operator < (TIntegral arg) const {
    return lt(arg);
  }

  // method: operator <=
  //
  boolean operator <= (TIntegral arg) const {
    return le(arg);
  }

  // method: operator >
  //
  boolean operator > (TIntegral arg) const {
    return gt(arg);
  }

  // method: operator >=
  //
  boolean operator >= (TIntegral arg) const {
    return ge(arg);
  }

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  equality and comparison methods
  //
  //---------------------------------------------------------------------------

  // method: eq
  //
  boolean eq(TIntegral arg) const {
    return (value_d == arg);
  }

  // method: eq
  //
  boolean eq(TIntegral arg1, TIntegral arg2) const {
    return (arg1 == arg2);
  }

  // method: ne
  //
  boolean ne(TIntegral arg) const {
    return (value_d != arg);
  }
  
  // method: lt
  //
  boolean lt(TIntegral arg) const {
    return (value_d < arg);
  }

  // method: le
  //
  boolean le(TIntegral arg) const {
    return (value_d <= arg);
  }

  // method: gt
  //
  boolean gt(TIntegral arg) const {
    return (value_d > arg);
  }

  // method: ge
  //
  boolean ge(TIntegral arg) const {
    return (value_d >= arg);
  }

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  bitwise methods
  //
  //--------------------------------------------------------------------------

  // method: band
  //
  TIntegral band(TIntegral arg) {
    return band(value_d, arg);
  }

  // method: band
  //
  TIntegral band(TIntegral arg1, TIntegral arg2) {
    return (value_d = (arg1 & arg2));
  }

  // method: bor
  //
  TIntegral bor(TIntegral arg) {
    return bor(value_d, arg);
  }

  // method: bor
  //
  TIntegral bor(TIntegral arg1, TIntegral arg2) {
    return (value_d = (arg1 | arg2));
  }

  // method: bxor
  //
  TIntegral bxor(TIntegral arg) {
    return bxor(value_d, arg);
  }

  // method: bxor
  //
  TIntegral bxor(TIntegral arg1, TIntegral arg2) {
     return (value_d = (arg1 ^ arg2));
  }

  // method: brs
  //
  TIntegral brs(TIntegral incr) {
    return brs(value_d, incr);
  }

  // method: brs
  //
  TIntegral brs(TIntegral arg, TIntegral incr) {
    return (value_d = (arg >> incr));
  }

  // method: bls
  //
  TIntegral bls(TIntegral incr) {
    return bls(value_d, incr);
  }

  // method: bls
  //
  TIntegral bls(TIntegral arg, TIntegral incr) {
    return (value_d = (arg << incr));
  }

  // method: bcmpl
  //
  TIntegral bcmpl() {
    return bcmpl(value_d);
  }

  // method: bcmpl
  //
  TIntegral bcmpl(TIntegral arg) {
    return (value_d = (~arg));
  }
  
  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  basic mathematical methods
  //
  //---------------------------------------------------------------------------

  // method: add
  //
  TIntegral add(TIntegral arg2) {
    return add(value_d, arg2);
  }

  // method: add
  //
  TIntegral add(TIntegral arg1, TIntegral arg2) {
    return (value_d = arg1 + arg2);
  }

  // method: sub
  //
  TIntegral sub(TIntegral arg2) {
    return sub(value_d, arg2);
  }

  // method: sub
  //
  TIntegral sub(TIntegral arg1, TIntegral arg2) {
    return (value_d = arg1 - arg2);
  }

  // method: mult
  //
  TIntegral mult(TIntegral arg2) {
    return mult(value_d, arg2);
  }

  // method: mult
  //
  TIntegral mult(TIntegral arg1, TIntegral arg2) {
    return (value_d = arg1 * arg2);
  }

  // method: div
  //
  TIntegral div(TIntegral arg2) {
    return div(value_d, arg2);
  }

  // method: div
  //
  TIntegral div(TIntegral arg1, TIntegral arg2) {
    return (value_d = arg1 / arg2);
  }

  // method: mod
  //
  TIntegral mod(TIntegral arg2) {
    return mod(value_d, arg2);
  }

  // method: mod
  //
  TIntegral mod(TIntegral arg1, TIntegral arg2) {
    return (value_d = arg1 % arg2);
  }

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  other mathematical methods (listed alphabetically)
  //
  //--------------------------------------------------------------------------

  // method: abs
  //
  TIntegral abs() {
    return abs(value_d);
  }

  // method: abs
  //
  TIntegral abs(TIntegral arg) {
    if (arg > (TIntegral)0) {
      value_d = arg;
    }
    else {
      value_d = -arg;
    }
    return value_d;
  }
 
  // method: acos
  //
  TIntegral acos() {
    return acos(value_d);
  }

  // method: acos
  //
  TIntegral acos(TIntegral arg) {
    return (value_d = (TIntegral)Integral::acos(arg));
  }

  // method: acosh
  //
  TIntegral acosh() {
    return acosh(value_d);
  }

  // method: acosh
  //
  TIntegral acosh(TIntegral arg) {
    return (value_d = (TIntegral)Integral::acosh(arg));
  }

  // method: asin
  //
  TIntegral asin() {
    return asin(value_d);
  }

  // method: asin
  //
  TIntegral asin(TIntegral arg) {
    return (value_d = (TIntegral)Integral::asin(arg));
  }

  // method: asinh
  //
  TIntegral asinh() {
    return asinh(value_d);
  }

  // method: asinh
  //
  TIntegral asinh(TIntegral arg) {
    return (value_d = (TIntegral)Integral::asinh(arg));
  }

  // method: atan
  //
  TIntegral atan() {
    return atan(value_d);
  }

  // method: atan
  //
  TIntegral atan(TIntegral arg) {
    return (value_d = (TIntegral)Integral::atan(arg));
  }

  // method: atanh
  //
  TIntegral atanh() {
    return atanh(value_d);
  }

  // method: atanh
  //
  TIntegral atanh(TIntegral arg) {
    return (value_d = (TIntegral)Integral::atanh(arg));
  }

  // method: ceil
  //
  TIntegral ceil() {
    return ceil(value_d);
  }

  // method: ceil
  //
  template <class TAIntegral>
  TIntegral ceil(TAIntegral arg) {
    return (value_d = (TIntegral)Integral::ceil(arg));
  }

  // method: centerClip
  //
  TIntegral centerClip(TIntegral min, TIntegral max) {
    return MScalarMethods::centerClip(*this, min, max, value_d);
  }

  // method: centerClip
  //
  TIntegral centerClip(TIntegral min, TIntegral max, TIntegral val) {
    return MScalarMethods::centerClip(*this, min, max, val);
  }
  
  // method: cos
  //
  TIntegral cos() {
    return cos(value_d);
  }
  
  // method: cos
  //
  TIntegral cos(TIntegral arg) {
    return (value_d = (TIntegral)Integral::cos(arg));
  }

  // method: cosh
  //
  TIntegral cosh() {
    return cosh(value_d);
  }
  
  // method: cosh
  //
  TIntegral cosh(TIntegral arg) {
    return (value_d = (TIntegral)Integral::cosh(arg));
  }

  // method: exp
  //
  TIntegral exp() {
    return exp(value_d);
  }

  // method: exp
  //
  TIntegral exp(TIntegral arg) {
    return (value_d = presentValue(Integral::exp(arg)));
  }

  // method: exp2
  //
  TIntegral exp2() {
    return exp2(value_d);
  }

  // method: exp2
  //
  TIntegral exp2(TIntegral arg) {
    return (value_d = presentValue(Integral::exp2(arg)));
  }

  // method: exp10
  //
  TIntegral exp10() {
    return exp10(value_d);
  }

  // method: exp10
  //
  TIntegral exp10(TIntegral arg) {
    return (value_d =  presentValue(Integral::exp10(arg)));
  }
  
  // method: factorial
  //
  TIntegral factorial() {
    return factorial(value_d);
  }
  
  // method: factorial
  //
  TIntegral factorial(TIntegral arg) {
    value_d = 1;    
    for (TIntegral count = 2; count <= arg; count++) {
      value_d *= count;
    }
    return value_d;
  }

  // method: floor
  //
  TIntegral floor() {
    return floor(value_d);
  }

  // method: floor
  //
  template <class TAIntegral>
  TIntegral floor(TAIntegral arg) {
    return (value_d = (TIntegral)Integral::floor(arg));
  }

  // method: fraction
  //
  TIntegral fraction() {
    return fraction(value_d);
  }
  
  // method: fraction
  //
  TIntegral fraction(TIntegral arg) {
    return (value_d = (TIntegral)0);
  }
  
  // gaussian random number generation methods
  //
  TIntegral grand(TIntegral mean, TIntegral stddev,
		  Random& generator = Random::GLOBAL_GAUSSIAN);

  // method: hash
  //
  ulong hash(ulong hsize) const {
    return Integral::hash(value_d,  hsize);
  }  
  
  // method: inverse
  //
  TIntegral inverse() {
    return inverse(value_d);
  }
  
  // method: inverse
  //
  TIntegral inverse(TIntegral arg) {
    value_d = 1;
    return (value_d /= arg);
  }

  // method: limit
  //
  TIntegral limit(TIntegral min, TIntegral max) {
    return MScalarMethods::limit(*this, min, max, value_d);
  }
  
  // method: limit
  //
  TIntegral limit(TIntegral min, TIntegral max, TIntegral val) {
    return MScalarMethods::limit(*this, min, max, val);
  }
  
  // method: limitMag
  //
  TIntegral limitMag(TIntegral thresh, TIntegral output) {
    return MScalarMethods::limitMag(*this, thresh, output, value_d);
  }
  
  // method: limitMag
  //
  TIntegral limitMag(TIntegral thresh, TIntegral output, TIntegral val) {
    return MScalarMethods::limitMag(*this, thresh, output, val);
  }

  // method: log
  //
  TIntegral log() {
    return log(value_d);
  }
  
  // method: log
  //
  TIntegral log(TIntegral arg) {
    return (value_d = presentValue(Integral::log(arg)));
  }

  // method: log2
  //
  TIntegral log2() {
    return log2(value_d);
  }

  // method: log2
  //
  TIntegral log2(TIntegral arg) {
    return (value_d = presentValue(Integral::log2(arg)));
  }
  
  // method: log10
  //
  TIntegral log10() {
    return log10(value_d);
  }
  
  // method: log10
  //
  TIntegral log10(TIntegral arg) {
    return (value_d = presentValue(Integral::log10(arg)));
  }
   
  // method: log1p
  //
  TIntegral log1p() {
    return log1p(value_d);
  }
   
  // method: log1p
  //
  TIntegral log1p(TIntegral arg) {
    return (value_d = presentValue(Integral::log1p(arg)));
  }
   
  // method: max
  //
  TIntegral max(TIntegral arg) {
    return max(value_d, arg);
  }
 
  // method: max
  //
  TIntegral max(TIntegral arg_1, TIntegral arg_2) {
    if (arg_1 > arg_2) {
      return (value_d = arg_1);
    }
    return (value_d = arg_2);
  }
  
  // method: min
  //
  TIntegral min(TIntegral arg) {
    return min(value_d, arg);
  }
  
  // method: min
  //
  TIntegral min(TIntegral arg_1, TIntegral arg_2) {
    if (arg_1 > arg_2) {
      return (value_d = arg_2);
    }
    else {
      return (value_d = arg_1);    
    }
  }

  // method: neg
  //
  TIntegral neg() {
    return neg(value_d);
  }
  
  // method: neg
  //
  TIntegral neg(TIntegral arg) {
    return (value_d = -arg);
  }

  // method: pow
  //
  TIntegral pow(TIntegral exponent) {
    return pow(value_d, exponent);
  }
  
  // method: pow
  //
  TIntegral pow(TIntegral arg, TIntegral exponent) {
    return (value_d = presentValue(Integral::pow(arg, exponent)));
  }

  // random number generation methods
  //
  TIntegral rand(Random& generator = Random::GLOBAL_UNIFORM);
  TIntegral rand(TIntegral min, TIntegral max,
		 Random& generator = Random::GLOBAL_UNIFORM);
  
  // method: rceil
  //
  TIntegral rceil() {
    return rceil(value_d);
  }

  // method: rceil
  //
  TIntegral rceil(TIntegral arg) {
    return (value_d = (TIntegral)(Integral::ceil(arg) - arg));
  }

  // method: rfloor
  //
  TIntegral rfloor() {
    return rfloor(value_d);
  }

  // method: rfloor
  //
  TIntegral rfloor(TIntegral arg) {
    return (value_d = (TIntegral)(Integral::floor(arg) - arg));
  }

  // method: round
  //
  TIntegral round() {
    return round(value_d);
  }

  // method: round
  //
  template <class TAIntegral>
  TIntegral round(TAIntegral arg) {
    return (value_d = (TIntegral)Integral::round(arg));
  }

  // method: sign
  //
  TIntegral sign() {
    return sign(value_d);
  }
  
  // method: sign
  //
  TIntegral sign(TIntegral arg) {
    if (arg < (TIntegral)0.0) {
      return (value_d = (TIntegral)Integral::NEGATIVE);
    }
    return (value_d = (TIntegral)Integral::POSITIVE);    
  }
  
  // method: sin
  //
  TIntegral sin() {
    return sin(value_d);
  }

  // method: sin
  //
  TIntegral sin(TIntegral arg) {
    return (value_d = (TIntegral)Integral::sin(arg));
  }

  // method: sinh
  //
  TIntegral sinh() {
    return sinh(value_d);
  }

  // method: sinh
  //
  TIntegral sinh(TIntegral arg) {
    return (value_d = (TIntegral)Integral::sinh(arg));
  }

  // method: sqrt
  //
  TIntegral sqrt() {
    return sqrt(value_d);
  }
   
  // method: sqrt
  //
  TIntegral sqrt(TIntegral arg) {
    return (value_d = presentValue(Integral::sqrt(arg)));
  }

  // method: square
  //
  TIntegral square() {
    return (value_d * value_d);
  }
   
  // method: square
  //
  TIntegral square(TIntegral arg) {
    return (value_d = arg * arg);
  }

  // method: tan
  //
  TIntegral tan() {
    return tan(value_d);
  }

  // method: tan
  //
  TIntegral tan(TIntegral arg) {
    return (value_d = (TIntegral)Integral::tan(arg));
  }

  // method: tanh
  //
  TIntegral tanh() {
    return tanh(value_d);
  }

  // method: tanh
  //
  TIntegral tanh(TIntegral arg) {
    return (value_d = (TIntegral)Integral::tanh(arg));
  }

  //---------------------------------------------------------------------------
  //
  // private methods
  //
  //---------------------------------------------------------------------------
private:
  
  // friend functions:
  //  while friend functions are generally discouraged in the ISIP
  //  environment, they are used in the MScalar class to circumvent the
  //  template delayed compilation problem. 
  //
  friend class MScalarMethods;
};

// the presentValue method is used to construct values for
// mathematical methods which have approximate
// values for integer types and exact values for floating point types
//

template<> double
MScalar<double, float64>::presentValue(double arg) const {
  return (double)arg;
}

template<> float
MScalar<float, float32>::presentValue(double arg) const {
  return (float)arg;
}

template<> complexdouble
MScalar<complexdouble, float64>::presentValue(complexdouble arg) const {
  return (complexdouble)arg;
}

template<> complexfloat
MScalar<complexfloat, float32>::presentValue(complexdouble arg) const {
  return (complexfloat)arg;
}

template<> complexlong
MScalar<complexlong, int32>::presentValue(complexdouble arg) const {
  return (complexlong)Integral::round(arg);
}

// template specializations for mathematical methods which have type-specific
// values. typically, these are methods that have approximate
// values for integer types and exact values for floating point types.
//

template<> double
MScalar<double, float64>::mod(double arg1, double arg2) {
  return arg1;
}

template<> float
MScalar<float, float32>::mod(float arg1, float arg2) {
  return arg1;
}

template<> double
MScalar<double, float64>::operator%(double arg) const {
  return arg;
}

template<> float
MScalar<float, float32>::operator%(float arg) const {
  return arg;
}

template<> MScalar<double, float64>&
MScalar<double, float64>::operator%= (double arg) {
  return *this;
}

template<> MScalar<float, float32>&
MScalar<float, float32>::operator%= (float arg) {
  return *this;
}

template<> double
MScalar<double, float64>::bor(double arg1, double arg2) {
  return arg1;
}

template<> float
MScalar<float, float32>::bor(float arg1, float arg2) {
  return arg1;
}

template<> double
MScalar<double, float64>::band(double arg1, double arg2) {
  return arg1;
}

template<> float
MScalar<float, float32>::band(float arg1, float arg2) {
  return arg1;
}

template<> double
MScalar<double, float64>::bxor(double arg1, double arg2) {
  return arg1;
}

template<> float
MScalar<float, float32>::bxor(float arg1, float arg2) {
  return arg1;
}

template<> double
MScalar<double, float64>::brs(double arg1, double arg2) {
  return arg1;
}

template<> float
MScalar<float, float32>::brs(float arg1, float arg2) {
  return arg1;
}

template<> double
MScalar<double, float64>::bls(double arg1, double arg2) {
  return arg1;
}

template<> float
MScalar<float, float32>::bls(float arg1, float arg2) {
  return arg1;
}

template<> double
MScalar<double, float64>::bcmpl(double arg) {
  return arg;
}

template<> float
MScalar<float, float32>::bcmpl(float arg1) {
  return arg1;
}

template<> double
MScalar<double, float64>::fraction(double arg) {
  return (value_d = Integral::fraction(arg)); 
}

template<> float
MScalar<float, float32>::fraction(float arg) {
  return (value_d = Integral::fraction(arg)); 
}

template<> byte
MScalar<byte, byte8>::neg(byte arg) {
  Error::handle(name(), L"neg", Error::TEST, __FILE__, __LINE__);
  return (byte)0; 
}

template<> ulong
MScalar<ulong, uint32>::neg(ulong arg) {
    Error::handle(name(), L"neg", Error::TEST, __FILE__, __LINE__);
    return (ulong)0; 
}

template<> ullong
MScalar<ullong, uint64>::neg(ullong arg) {
   Error::handle(name(), L"neg", Error::TEST, __FILE__, __LINE__);
   return (ullong)0; 
}

template<> ushort
MScalar<ushort, uint16>::neg(ushort arg) {
   Error::handle(name(), L"neg", Error::TEST, __FILE__, __LINE__);
   return (ushort)0;
 }

template<> byte
MScalar<byte, byte8>::abs(byte arg) {
  Error::handle(name(), L"abs", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
 }

template<> ulong
MScalar<ulong, uint32>::abs(ulong arg) {
  Error::handle(name(), L"abs", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
 }

template<> ullong
MScalar<ullong, uint64>::abs(ullong arg) {
  Error::handle(name(), L"abs", Error::TEST, __FILE__, __LINE__);
  return (ullong)0; 
}

template<> ushort
MScalar<ushort, uint16>::abs(ushort arg) {
  Error::handle(name(), L"abs", Error::TEST, __FILE__, __LINE__);
  return (ushort)0; 
}

template<> byte
MScalar<byte, byte8>::sign(byte arg) {
    Error::handle(name(), L"sign", Error::TEST, __FILE__, __LINE__);
    return (byte)0;
}

template<> ulong
MScalar<ulong, uint32>::sign(ulong arg) {
  Error::handle(name(), L"sign", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ullong
MScalar<ullong, uint64>::sign(ullong arg) {
  Error::handle(name(), L"sign", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ushort
MScalar<ushort, uint16>::sign(ushort arg) {
  Error::handle(name(), L"sign", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> byte
MScalar<byte, byte8>::acos(byte arg) {
  Error::handle(name(), L"acos", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::acos(ullong arg) {
  Error::handle(name(), L"acos", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::acos(ulong arg) {
  Error::handle(name(), L"acos", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::acos(ushort arg) {
  Error::handle(name(), L"acos", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::acos(llong arg) {
  Error::handle(name(), L"acos", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::acos(long arg) {
  Error::handle(name(), L"acos", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::acos(short arg) {
  Error::handle(name(), L"acos", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}

template<> byte
MScalar<byte, byte8>::acosh(byte arg) {
  Error::handle(name(), L"acosh", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::acosh(ullong arg) {
  Error::handle(name(), L"acosh", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::acosh(ulong arg) {
  Error::handle(name(), L"acosh", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::acosh(ushort arg) {
  Error::handle(name(), L"acosh", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::acosh(llong arg) {
  Error::handle(name(), L"acosh", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::acosh(long arg) {
  Error::handle(name(), L"acosh", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::acosh(short arg) {
  Error::handle(name(), L"acosh", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}

template<> byte
MScalar<byte, byte8>::asin(byte arg) {
  Error::handle(name(), L"asin", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::asin(ullong arg) {
  Error::handle(name(), L"asin", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::asin(ulong arg) {
  Error::handle(name(), L"asin", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::asin(ushort arg) {
  Error::handle(name(), L"asin", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::asin(llong arg) {
  Error::handle(name(), L"asin", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::asin(long arg) {
  Error::handle(name(), L"asin", Error::TEST, __FILE__, __LINE__);
  return (long)0; 
}
template<> short
MScalar<short, int16>::asin(short arg) {
  Error::handle(name(), L"asin", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}

template<> byte
MScalar<byte, byte8>::asinh(byte arg) {
  Error::handle(name(), L"asinh", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::asinh(ullong arg) {
  Error::handle(name(), L"asinh", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::asinh(ulong arg) {
  Error::handle(name(), L"asinh", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::asinh(ushort arg) {
  Error::handle(name(), L"asinh", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::asinh(llong arg) {
  Error::handle(name(), L"asinh", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::asinh(long arg) {
  Error::handle(name(), L"asinh", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::asinh(short arg) {
  Error::handle(name(), L"asinh", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}

template<> byte
MScalar<byte, byte8>::atan(byte arg) {
  Error::handle(name(), L"atan", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::atan(ullong arg) {
  Error::handle(name(), L"atan", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::atan(ulong arg) {
  Error::handle(name(), L"atan", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::atan(ushort arg) {
  Error::handle(name(), L"atan", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::atan(llong arg) {
  Error::handle(name(), L"atan", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::atan(long arg) {
  Error::handle(name(), L"atan", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::atan(short arg) {
  Error::handle(name(), L"atan", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}

template<> byte
MScalar<byte, byte8>::atanh(byte arg) {
  Error::handle(name(), L"atanh", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::atanh(ullong arg) {
  Error::handle(name(), L"atanh", Error::TEST, __FILE__, __LINE__);;
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::atanh(ulong arg) {
  Error::handle(name(), L"atanh", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::atanh(ushort arg) {
  Error::handle(name(), L"atanh", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::atanh(llong arg) {
  Error::handle(name(), L"atanh", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::atanh(long arg) {
  Error::handle(name(), L"atanh", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::atanh(short arg) {
  Error::handle(name(), L"atanh", Error::TEST, __FILE__, __LINE__);
  return (short)0; 
}

template<> byte
MScalar<byte, byte8>::cos(byte arg) {
  Error::handle(name(), L"cos", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::cos(ullong arg) {
  Error::handle(name(), L"cos", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::cos(ulong arg) {
  Error::handle(name(), L"cos", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::cos(ushort arg) {
  Error::handle(name(), L"cos", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::cos(llong arg) {
  Error::handle(name(), L"cos", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::cos(long arg) {
  Error::handle(name(), L"cos", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::cos(short arg) {
  Error::handle(name(), L"cos", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}

template<> byte
MScalar<byte, byte8>::cosh(byte arg) {
  Error::handle(name(), L"cosh", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::cosh(ullong arg) {
  Error::handle(name(), L"cosh", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::cosh(ulong arg) {
  Error::handle(name(), L"cosh", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::cosh(ushort arg) {
  Error::handle(name(), L"cosh", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::cosh(llong arg) {
  Error::handle(name(), L"cosh", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::cosh(long arg) {
  Error::handle(name(), L"cosh", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::cosh(short arg) {
  Error::handle(name(), L"cosh", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}
/*
template<> byte
MScalar<byte, byte8>::floor(byte arg) {
  Error::handle(name(), L"floor", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::floor(ullong arg) {
  Error::handle(name(), L"floor", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::floor(ulong arg) {
  Error::handle(name(), L"floor", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::floor(ushort arg) {
  Error::handle(name(), L"floor", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::floor(llong arg) {
  Error::handle(name(), L"floor", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::floor(long arg) {
  Error::handle(name(), L"floor", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::floor(short arg) {
  Error::handle(name(), L"floor", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}
*/
template<> byte
MScalar<byte, byte8>::sin(byte arg) {
  Error::handle(name(), L"sin", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::sin(ullong arg) {
  Error::handle(name(), L"sin", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::sin(ulong arg) {
  Error::handle(name(), L"sin", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::sin(ushort arg) {
  Error::handle(name(), L"sin", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::sin(llong arg) {
  Error::handle(name(), L"sin", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::sin(long arg) {
  Error::handle(name(), L"sin", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::sin(short arg) {
  Error::handle(name(), L"sin", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}

template<> byte
MScalar<byte, byte8>::sinh(byte arg) {
  Error::handle(name(), L"sinh", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::sinh(ullong arg) {
  Error::handle(name(), L"sinh", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::sinh(ulong arg) {
  Error::handle(name(), L"sinh", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::sinh(ushort arg) {
  Error::handle(name(), L"sinh", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::sinh(llong arg) {
  Error::handle(name(), L"sinh", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::sinh(long arg) {
  Error::handle(name(), L"sinh", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::sinh(short arg) {
  Error::handle(name(), L"sinh", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}

template<> byte
MScalar<byte, byte8>::tan(byte arg) {
  Error::handle(name(), L"tan", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::tan(ullong arg) {
  Error::handle(name(), L"tan", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::tan(ulong arg) {
  Error::handle(name(), L"tan", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::tan(ushort arg) {
  Error::handle(name(), L"tan", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::tan(llong arg) {
  Error::handle(name(), L"tan", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::tan(long arg) {
  Error::handle(name(), L"tan", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}


template<> short
MScalar<short, int16>::tan(short arg) {
  Error::handle(name(), L"tan", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}

template<> byte
MScalar<byte, byte8>::tanh(byte arg) {
  Error::handle(name(), L"tanh", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::tanh(ullong arg) {
  Error::handle(name(), L"tanh", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::tanh(ulong arg) {
  Error::handle(name(), L"tanh", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::tanh(ushort arg) {
  Error::handle(name(), L"tanh", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::tanh(llong arg) {
  Error::handle(name(), L"tanh", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::tanh(long arg) {
  Error::handle(name(), L"tanh", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::tanh(short arg) {
  Error::handle(name(), L"tanh", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}
/*
template<> byte
MScalar<byte, byte8>::ceil(byte arg) {
  Error::handle(name(), L"ceil", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::ceil(ullong arg) {
  Error::handle(name(), L"ceil", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::ceil(ulong arg) {
  Error::handle(name(), L"ceil", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::ceil(ushort arg) {
  Error::handle(name(), L"ceil", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::ceil(llong arg) {
  Error::handle(name(), L"ceil", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::ceil(long arg) {
  Error::handle(name(), L"ceil", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::ceil(short arg) {
  Error::handle(name(), L"ceil", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}
*/
template<> byte
MScalar<byte, byte8>::rceil(byte arg) {
  Error::handle(name(), L"rceil", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::rceil(ullong arg) {
  Error::handle(name(), L"rceil", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::rceil(ulong arg) {
  Error::handle(name(), L"rceil", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::rceil(ushort arg) {
  Error::handle(name(), L"rceil", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::rceil(llong arg) {
  Error::handle(name(), L"rceil", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::rceil(long arg) {
  Error::handle(name(), L"rceil", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}

template<> short
MScalar<short, int16>::rceil(short arg) {
  Error::handle(name(), L"rceil", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}

template<> byte
MScalar<byte, byte8>::rfloor(byte arg) {
  Error::handle(name(), L"rfloor", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::rfloor(ullong arg) {
  Error::handle(name(), L"rfloor", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::rfloor(ulong arg) {
  Error::handle(name(), L"rfloor", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::rfloor(ushort arg) {
  Error::handle(name(), L"rfloor", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::rfloor(llong arg) {
  Error::handle(name(), L"rfloor", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::rfloor(long arg) {
  Error::handle(name(), L"rfloor", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}


template<> short
MScalar<short, int16>::rfloor(short arg) {
  Error::handle(name(), L"rfloor", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}
/*
template<> byte
MScalar<byte, byte8>::round(byte arg) {
  Error::handle(name(), L"round", Error::TEST, __FILE__, __LINE__);
  return (byte)0;
}

template<> ullong
MScalar<ullong, uint64>::round(ullong arg) {
  Error::handle(name(), L"round", Error::TEST, __FILE__, __LINE__);
  return (ullong)0;
}

template<> ulong
MScalar<ulong, uint32>::round(ulong arg) {
  Error::handle(name(), L"round", Error::TEST, __FILE__, __LINE__);
  return (ulong)0;
}

template<> ushort
MScalar<ushort, uint16>::round(ushort arg) {
  Error::handle(name(), L"round", Error::TEST, __FILE__, __LINE__);
  return (ushort)0;
}

template<> llong
MScalar<llong, int64>::round(llong arg) {
  Error::handle(name(), L"round", Error::TEST, __FILE__, __LINE__);
  return (llong)0;
}

template<> long
MScalar<long, int32>::round(long arg) {
  Error::handle(name(), L"round", Error::TEST, __FILE__, __LINE__);
  return (long)0;
}


template<> short
MScalar<short, int16>::round(short arg) {
  Error::handle(name(), L"round", Error::TEST, __FILE__, __LINE__);
  return (short)0;
}
*/
  
// end of include file
//
#endif
