// file: $isip/class/numeric/Histogram/hist_06.cc
// version: $Id: hist_06.cc,v 1.2 2001/12/26 22:32:50 alphonso Exp $
//

// isip include files
//
#include "Histogram.h"

// method: setBins
//
// arguments:
//  const TVector& bins: (input) input binscomputed result
//
// return: boolean value indicating status
//
// this method sets the bins to be used for histograms. the counts are reset
// to zero
//
template <class TVector>
boolean Histogram::setBins(const TVector& bins_a) {

  // assign the bins and reset the counts
  //
  bins_d.assign(bins_a);
  counts_d.setLength(bins_d.length(), false);
  counts_d.assign(0.0);

  // exit gracefully
  //
  return true;
}

// explicit instantiations
//
template boolean
Histogram::setBins<VectorByte>(const VectorByte&);
template boolean
Histogram::setBins<VectorDouble>(const VectorDouble&);
template boolean
Histogram::setBins<VectorFloat>(const VectorFloat&);
template boolean
Histogram::setBins<VectorLlong>(const VectorLlong&);
template boolean
Histogram::setBins<VectorLong>(const VectorLong&);
template boolean
Histogram::setBins<VectorShort>(const VectorShort&);
template boolean
Histogram::setBins<VectorUllong>(const VectorUllong&);
template boolean
Histogram::setBins<VectorUlong>(const VectorUlong&);
template boolean
Histogram::setBins<VectorUshort>(const VectorUshort&);

// method: setBins
//
// arguments:
//  TIntegral min: (input) minimum of the histogram range
//  TIntegral max: (input) maximum of the histogram range
//  long num_bins: (input) number of bins to use
//
// return: boolean value indicating status
//
// this method computes the bin locations from the min, max, num_bins
// and binning mode
//
// if mode_d == CENTERS then the bin width will be (max - min) / num_bins
//    and the bin centers will be (min + (i * bin_width) + bin_width / 2) for
//    0 <= i < num_bins
//
// if mode_d == EDGES then num_bins must be >= 2. the bin width will be
//    (max - min) / (num_bins - 1). The bin edges will be
//    (min + i * bin_width) for 0 <= i < num_bins
//
template <class T1, class T2>
boolean Histogram::setBins(T1 min_a, T2 max_a, long num_bins_a) {

  // error checking
  //
  if ((num_bins_a <= 0) || (max_a < min_a)) {
    return Error::handle(name(), L"setBins", Error::ARG, __FILE__, __LINE__);
  }
  
  // set the number of bins and initialize the count vector
  //
  bins_d.setLength(num_bins_a, false);
  counts_d.setLength(num_bins_a, false);
  counts_d.assign(0.0);
  
  // which mode are we computing for?
  //
  if (mode_d == CENTERS) {

    // determine the bin width
    //
    double bin_width = (double)(max_a - min_a) / (double)num_bins_a;
    double first = min_a + bin_width / 2;
    
    // loop and add to get the bin centers
    //
    for (long i = 0; i < num_bins_a; i++) {
      bins_d(i) = first + i * bin_width;
    }
  }
  else if (mode_d == EDGES) {
    
    // determine the bin width
    //
    double bin_width = (double)(max_a - min_a) / ((double)num_bins_a - 1);
    
    // loop and add to get the bin edges
    //
    for (long i = 0; i < num_bins_a; i++) {
      bins_d(i) = min_a + i * bin_width;
    }

    // set the last bin
    //
    bins_d(num_bins_a - 1) = max_a;
  }

  // exit gracefully
  //
  return true;
}

// explicit instantiations
//
template boolean
Histogram::setBins<byte, byte>(byte, byte, long);
template boolean
Histogram::setBins<double, double>(double, double, long);
template boolean
Histogram::setBins<float, float>(float, float, long);
template boolean
Histogram::setBins<llong, llong>(llong, llong, long);
template boolean
Histogram::setBins<long, long>(long, long, long);
template boolean
Histogram::setBins<short, short>(short, short, long);
template boolean
Histogram::setBins<ullong, ullong>(ullong, ullong, long);
template boolean
Histogram::setBins<ulong, ulong>(ulong, ulong, long);
template boolean
Histogram::setBins<ushort, ushort>(ushort, ushort, long);


// method: setCounts
//
// arguments:
//  const TVector& counts: (input) counts to initialize histogram with
//
// return: boolean value indicating status
//
// this method sets the counts vector. the input vector must be the same
// length as the internal bin vector
//
template <class TVector>
boolean Histogram::setCounts(const TVector& counts_a) {

  // verify the input vector length
  //
  if (counts_a.length() != bins_d.length()) {
    return Error::handle(name(), L"setCounts", Error::ARG, __FILE__, __LINE__);
  }

  // assign the counts vector
  //
  return (counts_d.assign(counts_a));
}

// explicit instantiations
//
template boolean
Histogram::setCounts<VectorByte>(const VectorByte&);
template boolean
Histogram::setCounts<VectorDouble>(const VectorDouble&);
template boolean
Histogram::setCounts<VectorFloat>(const VectorFloat&);
template boolean
Histogram::setCounts<VectorLlong>(const VectorLlong&);
template boolean
Histogram::setCounts<VectorLong>(const VectorLong&);
template boolean
Histogram::setCounts<VectorShort>(const VectorShort&);
template boolean
Histogram::setCounts<VectorUllong>(const VectorUllong&);
template boolean
Histogram::setCounts<VectorUlong>(const VectorUlong&);
template boolean
Histogram::setCounts<VectorUshort>(const VectorUshort&);
